#!/bin/bash

###* Basic Configuration
LC_ALL=C

app_file="${0}"
app_name="$(basename $0)"

real_home="${HOME}"
if [ "${SUDO_USER}" ]; then
    real_home="/home/${SUDO_USER}"
fi

opt_info_func_list=("info_basic")
opt_syslog_include=()

###* Help Functions
msg() {
    local msg="${1}"; shift
    printf "==> ${msg}\n" ${@}
}

msg2() {
    local msg=$1; shift
    printf "  -> ${msg}\n" "$@" >&2
}

msg_title() {
    local msg="${1}"; shift
    printf "\n# ${msg}\n\n" ${@}
}

msg_code() {
    printf "%s" "${@}" | sed 's/^/    /'
    printf "\n"
}

get_self_funcs() {
    grep -o "^${1}.*()" "${app_file}" | sed "s/^\(.*\)()/\1/" | sort
}

get_module_funcs() {
    # use echo to join lines
    echo $(get_self_funcs "module_") | sed 's/module_//g'
}

get_info_funcs() {
    # use echo to join lines
    echo $(get_self_funcs "info_") | sed 's/info_//g'
}

grep_block() {
    local keyword="$1"; shift
    local files=$@
    awk -v keyword="${keyword}" 'BEGIN{RS="\n\n"; n=0} $0 ~ keyword{print ""; print; n++} END{print n, "of", NR, "matched", keyword}' $files
}

# collect_file module files...
collect_file() {
    local module="${1}"; shift
    for f in ${@}; do
        do_collect_file "${module}" "${f}"
    done
}
do_collect_file() {
    local module="${1}"; shift
    local child_dest_dir="${dest_dir}/${module}/$(dirname "${1}")"
    mkdir -p "${child_dest_dir}"
    cp -fR -t "${child_dest_dir}" "$(echo "${1}" | sed "s#^~#${real_home}#")"
}

###* Functions to Get Certain System Information
info_basic() {
    msg_title "Computer Model"
    msg_code "$(do_info_basic_computer_model)"

    msg_title "Linux Release"
    msg_code "$(lsb_release -a 2>&1)"

    msg_title "Linux Kernel"
    msg_code "$(uname -a)"

    msg_title "Installed Deepin Packages"
    msg_code "$(dpkg -l | grep -i -e 'deepin-' -e 'dde-')"
}
do_info_basic_computer_model() {
    # need root permission
    for f in /sys/class/dmi/id/*; do
        if [ -f "${f}" -a -r "${f}" ]; then
            printf "%s:\t%s\n" $f "$(cat $f)";
        fi
    done
}

info_service() {
    initctl list | column -t
}

info_env() {
    env
}

info_package() {
    dpkg -l
}

info_device() {
    msg_title "CPU"
    msg_code "$(cat /proc/cpuinfo)"

    msg_title "Memory"
    msg_code "$(free -t -h)"
    msg_code "$(cat /proc/meminfo)"

    msg_title "USB Devices"
    msg_code "$(lsusb)"

    msg_title "PCI Devices"
    msg_code "$(lspci -vvnn)"

    # need root permission
    msg_title "Hardware Lister(lshw)"
    msg_code "$(lshw 2>/dev/null)"
}

info_driver() {
    msg_title "Loaded Drivers"
    msg_code "$(lsmod)"

    msg_title "Driver Modules File"
    msg_code "$(cat /etc/modules)"

    msg_title "Driver Blacklist File"
    msg_code "$(do_info_driver_blacklist)"

    msg_title "Installed Driver Packages"
    msg_code "$(dpkg -l | grep -e driver -e catalyst -e nvidia)"
}
do_info_driver_blacklist() {
    for f in /etc/modprobe.d/*; do
        printf "${f}\n"
    done
    printf "\n\n"
    for f in /etc/modprobe.d/*; do
        printf "\n## ${f}\n"
        cat "${f}"
    done
}

info_kernel() {
    cat /var/log/dmesg
}

info_video() {
    msg_title "Video Devices"
    msg_code "$(lspci -vvnn | grep_block 'VGA ')"

    msg_title "Video Driver Packages"
    msg_code "$(dpkg -l | grep -e xorg-video -e catalyst -e nvidia)"
}

info_network() {
    msg_title "Network Devices"
    if ! is_info_device_will_run; then
        msg_code "$(lshw -C network 2>/dev/null)"
    fi
    msg_code "$(lspci -vvnn | grep_block '[nN]etwork')"
    msg_code "$(lsusb -v 2>/dev/null | grep_block '[nN]et')"

    msg_title "Network Status"
    msg_code "$(ifconfig -v -a)"
    msg_code "$(iwconfig 2>/dev/null)"

    msg_title "NetworkManager State"
    msg_code "$(nmcli nm status)"
    msg_code "$(nmcli con list)"
    msg_code "$(nm-tool)"

    msg_title "ModemManager State"
    msg_code "$(mmcli -L)"

    msg_title "Wireless Device Switches(rfkill)"
    msg_code "$(rfkill list all)"

    msg_title "Wireless Access Points"
    msg_code "$(iwlist scan 2>/dev/null)"

    msg_title "Network Interface File"
    msg_code "$(cat /etc/network/interfaces)"

    msg_title "DNS Configuration(resolv.conf)"
    msg_code "$(cat /etc/resolv.conf)"

    msg_title "Route Table"
    msg_code "$(route)"
}

info_bluetooth() {
    msg_title "Bluetooth Devices"
    msg_code "$(hciconfig -a)"
    msg_code "$(lspci -vvnn | grep_block '[bB]luetooth')"
    msg_code "$(lsusb | grep -i bluetooth)"

    msg_title "Loaded Bluetooth Drivers"
    msg_code "$(lsmod | grep -e btusb -e bluetooth -e hidp -e rfcomm)"

    msg_title "Wireless Device Switches(rfkill)"
    msg_code "$(rfkill list all)"
}

info_disk() {
    # need root permission
    if ! is_info_device_will_run; then
        msg_title "Disk Devices"
        msg_code "$(lshw -C disk -C storage 2>/dev/null)"
    fi

    msg_title "Disk Partition Table"
    msg_code "$(lsblk)"
}

info_syslog() {
    if [ ${#opt_syslog_include[@]} -gt 0 ]; then
        cat /var/log/syslog{.1,} 2>/dev/null | grep -i ${opt_syslog_include[@]}
    else
        cat /var/log/syslog{.1,} 2>/dev/null
    fi
}
include_syslog_keyword() {
    local len=${#opt_syslog_include[@]}
    opt_syslog_include[$len]="-e"
    ((len++))
    opt_syslog_include[$len]="${1}"
}

include_info() {
    local len=${#opt_info_func_list[@]}
    opt_info_func_list[$len]="info_${1}"
}
is_info_device_will_run() {
    for f in ${opt_info_func_list[@]}; do
        if [ "${f}" = "info_device" ]; then
            return 0
        fi
    done
    return 1
}
exec_info_funcs() {
    for f in ${opt_info_func_list[@]}; do
        msg2 "executing ${f}..."
        case "${f}" in
            "info_service") "${f}" >> "${file_service}";;
            "info_env")     "${f}" >> "${file_env}";;
            "info_package") "${f}" >> "${file_package}";;
            "info_device")  "${f}" >> "${file_device}";;
            "info_driver")  "${f}" >> "${file_driver}";;
            "info_kernel")  "${f}" >> "${file_kernel}";;
            "info_syslog")  "${f}" >> "${file_syslog}";;
            *)              "${f}" >> "${file_master}";;
        esac
    done
}

###* Modules
module_all() {
    # clean predefined options and execute all functions that could
    # get system information
    for f in $(get_module_funcs); do
        if [ "${f}" != "all" ]; then
            module_"${f}"
        fi
    done

    opt_info_func_list=()
    for f in $(get_info_funcs); do
        include_info "${f}"
    done
}

module_background() {
    include_info "syslog"
    include_syslog_keyword "com.deepin.SessionManager"
    include_syslog_keyword "com.deepin.daemon.ThemeManager"
    msg_title "GSettings Key-Values for Background" >> "${file_gsettings}"
    msg_code "$(gsettings list-recursively com.deepin.dde.personalization)"  >> "${file_gsettings}"
}

module_bluetooth() {
    include_info "bluetooth"
    include_info "kernel"
    include_info "driver"
    include_info "syslog"
    include_syslog_keyword "com.deepin.daemon.Bluetooth"
    include_syslog_keyword "bluetooth"
    collect_file "bluetooth" "~/.config/deepin/bluetooth.json"
}

module_bootmgr() {
    include_info "disk"
    include_info "syslog"
    include_syslog_keyword "com.deepin.daemon.Grub2"
    include_syslog_keyword "com.deepin.daemon.Grub2Ext"
    collect_file "bootmgr" /boot/grub/grub.cfg
    collect_file "bootmgr" /etc/default/grub
    collect_file "bootmgr" /var/cache/deepin/grub2.json
}

module_display() {
    include_info "video"
    include_info "syslog"
    include_syslog_keyword "com.deepin.SessionManager"
    include_syslog_keyword "com.deepin.daemon.Display"
    collect_file "display" "~/.config/deepin_monitors.json"
}

module_desktop() {
    collect_file "desktop" "/tmp/dde-desktop.log"
}

module_dock() {
    include_info "syslog"
    include_syslog_keyword "com.deepin.daemon.Dock"
    collect_file "dock" "/tmp/dde-dock.log"
}

module_launcher() {
    include_info "syslog"
    include_syslog_keyword "dde-daemon/launcher-daemon"
    collect_file "launcher" "/tmp/dde-launcher.log"
}

module_login() {
    include_info "syslog"
    include_syslog_keyword "com.deepin.SessionManager"
    include_syslog_keyword "com.deepin.daemon.Display"
    collect_file "login" "~/.xsession-errors"
    collect_file "login" "/var/log/lightdm" # need root permission
}

module_network() {
    include_info "network"
    include_info "driver"
    include_info "syslog"
    include_info "env"
    include_syslog_keyword "com.deepin.daemon.Network"
    include_syslog_keyword "NetworkManager"
    include_syslog_keyword "ModemManager"
    include_syslog_keyword "wpa_supplicant"
    include_syslog_keyword "dhclient"
    include_syslog_keyword "dnsmasq"
    include_syslog_keyword "avahi-daemon"
    collect_file "network" "~/.config/deepin/network.json"
}

###* Main
arg_module="all"              # if no arguments, just execute rules in module_all()
arg_info_type=
arg_complete_opt=
arg_help=

show_usage() {
    cat <<EOF
${app_name} [-p <info_type>] [-h] [<module>]
options:
    -p, print certain types of system information, the type coulde be:
        $(get_info_funcs)
    -h, show this message

    if there is no other options, will collect debug information for
    special module and save to archive file in current directory,
    the module could be (default: all):
        $(get_module_funcs)
EOF
    exit 1
}

# dispatch arguments
while [ ${#} -gt 0 ]; do
    case ${1} in
        -p) arg_info_type="${2}"; arg_module=""; shift; shift; break;;
        -C) arg_complete_opt="${2}"; shift; shift; break;;
        -h) arg_help=t; break;;
        *)  arg_module="${1}"; shift; break;;
    esac
done

if [ "${arg_help}" ]; then
    show_usage
fi

if [ "${arg_complete_opt}" ]; then
    case "${arg_complete_opt}" in
        -p) get_info_funcs;;
        *)  get_module_funcs;;
    esac
    exit
fi

if [ "${arg_info_type}" ]; then
    case "${arg_info_type}" in
        syslog)
            for a in ${@}; do
                include_syslog_keyword "${a}"
            done;;
    esac
    info_"${arg_info_type}"
    exit
fi

if [ "${arg_module}" ]; then
    # ensure this script running as root
    if [ ! "${UID}" -eq 0 ]; then
        printf "please run ${app_name} as root, just use: sudo ${app_name}"
        exit 1
    fi

    # global variables
    result_tag="${app_name}-results-${arg_module}-$(date "+%Y%m%d-%s")"
    dest_dir="/tmp/${result_tag}"
    result_archive="${result_tag}.tar.gz"
    file_master="${dest_dir}/sysinfo.md"
    file_syslog="${dest_dir}/syslog"
    file_service="${dest_dir}/services"
    file_env="${dest_dir}/env"
    file_package="${dest_dir}/packages"
    file_device="${dest_dir}/devices.md"
    file_driver="${dest_dir}/drivers.md"
    file_kernel="${dest_dir}/dmesg"
    file_gsettings="${dest_dir}/gsettings.md"

    msg "Collecting system information and will take several seconds..."

    # prepare
    msg "Preparing temporary folder ${dest_dir}..."
    rm -rf "${dest_dir}"
    mkdir -p "${dest_dir}"

    # execute rules of module
    msg "Execute rules of module '${arg_module}'..."
    if ! module_"${arg_module}"; then
        exit 1
    fi
    exec_info_funcs

    # archive files
    msg "Archive files..."
    dest_files=$(ls -1 "${dest_dir}" | tr "\n" " ")
    tar -cvzf "${result_archive}" -C "${dest_dir}" ${dest_files}

    # split archive file if need
    archive_size="$(stat -c "%s" "${result_archive}")"
    max_file_size="5242880" # 5mb
    if [ "${archive_size}" -ge "${max_file_size}" ]; then
        msg "Split archive file..."
        split -b 5m -d "${result_archive}" "${result_archive}".part.
    fi

    msg "Finished, please report a bug to Deepin team with ${result_archive} in current folder as an attachment:"
    msg2 "http://www.linuxdeepin.com/mantis/bug_report_page.php"
    exit
fi



# Local Variables:
# mode: sh
# mode: orgstruct
# orgstruct-heading-prefix-regexp: "^\s*###"
# End:
